class MnistModel {
    static instance = new MnistModel()
    duringTraining = false
    trainStartTime = null
    datasetNotFound = false
    #modelPath = null
    #model = null
    #single = null
    #modelOutput = null
    #trainCallbacks = new Set()

    /**
     *   config = {
     *       modelPath: "...",
     *       trainDatasetPath: "...",
     *       testDatasetPath: "...",
     *   }
     */
    async open(config) {
        this.#modelPath = config.modelPath
        this.#model = tizen.ml.trainer.createModel()

        try {
            this.#model.load(this.#modelPath, "FORMAT_INI_WITH_BIN")
        }
        catch(error) {
            Logger.instance.error(error)
            this.#buildModel(this.#model)
            Logger.instance.debug("Model built")
        }

        if(tizen.filesystem.pathExists(config.trainDatasetPath) &&
           tizen.filesystem.isFile(config.trainDatasetPath) &&
           tizen.filesystem.pathExists(config.testDatasetPath) &&
           tizen.filesystem.isFile(config.testDatasetPath)) {
            const dataset = tizen.ml.trainer.createFileDataset(
                config.trainDatasetPath,
                null,
                config.testDatasetPath
            )
            this.#model.setDataset(dataset)
            const optimizer = tizen.ml.trainer.createOptimizer("OPTIMIZER_ADAM")
            this.#model.setOptimizer(optimizer)
            this.#model.compile({loss: "cross"})
            await this.#saveModel()
            this.#openSingle()
            Logger.instance.debug(this.#model.summarize())
        }
        else {
            this.datasetNotFound = true
            Logger.instance.error("Dataset not found")
        }
    }

    train(epochs = 1) {
        if(this.duringTraining) {
            return
        }

        if(this.#single) {
            this.#single.close()
            this.#single = null
            this.#modelOutput.dispose()
            this.#modelOutput = null
        }

        Logger.instance.info("Training model")
        this.duringTraining = true
        this.trainStartTime = Date.now()
        this.#model.run({
            epochs: epochs,
            batch_size: 32,
        }, async () => {
            Logger.instance.info("Model training done")
            this.duringTraining = false
            this.trainStartTime = null
            await this.#saveModel()
            this.#openSingle()
            this.#trainCallbacks.forEach(c => c())
        }, error => {
            Logger.instance.error(`Error training model: ${error}`)
            this.duringTraining = false
            this.trainStartTime = null
            this.#trainCallbacks.forEach(c => c(error))
        })
    }

    predict(x) {
        return new Promise((resolve, reject) => {
            if(!this.#single) {
                return resolve(null)
            }

            this.#single.invokeAsync(x, y => {
                resolve(y)
            }, error => {
                reject(error)
            })
        })
    }

    addTrainEventListener(callback) {
        this.#trainCallbacks.add(callback)
    }

    removeTrainEventListener(callback) {
        this.#trainCallbacks.delete(callback)
    }

    #openSingle() {
        if(this.#single) {
            this.#single.close()
        }

        if(!this.#modelOutput) {
            this.#modelOutput = new tizen.ml.TensorsInfo()
            this.#modelOutput.addTensorInfo("output", "FLOAT32", [10])
        }

        this.#single = tizen.ml.single.openModel(this.#modelPath, null, this.#modelOutput, "NNTR_INF", "AUTO")
    }

    #buildModel(model) {
        const layer0 = tizen.ml.trainer.createLayer("LAYER_INPUT")
        layer0.setProperty("input_shape", "1:28:28")
        model.addLayer(layer0)

        const layer1 = tizen.ml.trainer.createLayer("LAYER_CONV2D")
        layer1.setProperty("filters", "32")
        layer1.setProperty("kernel_size", "5, 5")
        layer1.setProperty("activation", "relu")
        model.addLayer(layer1)

        const layer2 = tizen.ml.trainer.createLayer("LAYER_POOLING2D")
        layer2.setProperty("pooling", "max")
        layer2.setProperty("pool_size", "2, 2")
        layer2.setProperty("stride", "2, 2")
        model.addLayer(layer2)

        const layer3 = tizen.ml.trainer.createLayer("LAYER_CONV2D")
        layer3.setProperty("filters", "64")
        layer3.setProperty("kernel_size", "5, 5")
        layer3.setProperty("activation", "relu")
        model.addLayer(layer3)

        const layer4 = tizen.ml.trainer.createLayer("LAYER_POOLING2D")
        layer4.setProperty("pooling", "max")
        layer4.setProperty("pool_size", "2, 2")
        layer4.setProperty("stride", "2, 2")
        model.addLayer(layer4)

        const layer5 = tizen.ml.trainer.createLayer("LAYER_FLATTEN")
        model.addLayer(layer5)

        const layer6 = tizen.ml.trainer.createLayer("LAYER_FC")
        layer6.setProperty("unit", "256")
        layer6.setProperty("activation", "relu")
        model.addLayer(layer6)

        const layer7 = tizen.ml.trainer.createLayer("LAYER_FC")
        layer7.setProperty("unit", "10")
        layer7.setProperty("activation", "softmax")
        model.addLayer(layer7)
    }

    #saveModel() {
        return new Promise((resolve, reject) => {
            const save = () => {
                Logger.instance.debug("saving model")
                try {
                    this.#model.saveToFile(this.#modelPath, "FORMAT_INI_WITH_BIN")
                    resolve()
                }
                catch(error) {
                    Logger.instance.error(error)
                    reject(error)
                }
            }
            tizen.filesystem.deleteFile(this.#modelPath, () => {
                tizen.filesystem.deleteFile(this.#modelPath.replace(".ini", ".bin"), save, save)
            }, save)
        })
    }
}
